import pandas as pd
import numpy as np
import idelib

mm_to_m = 1e-3


def get_columns(df, startswith_str):
    """
    Returns list of columns in data-frame `df` which begin with the
    `startswith_str` string.  The list is sorted prior to return.  This ensures
    the return order will be X->Y->Z for 3-axis sensors.
    """
    cols = [col for col in df if col.startswith(startswith_str)]
    cols.sort()
    return cols


def get_accel_cols(df, startswith_str='Accel', verbose=False):
    cols = get_columns(df, startswith_str)
    if verbose:
        print('Accel Columns Found:\n\t'+'\t'.join(['%s' % c for c in cols]))
    return cols


def get_gyro_cols(df, startswith_str='AngleRate', verbose=False):
    cols = get_columns(df, startswith_str)
    if verbose:
        print('Gyro Columns Found:\n\t'+'\t'.join(['%s' % c for c in cols]))
    return cols


def get_mag_cols(df, startswith_str='MagField', verbose=False):
    cols = get_columns(df, startswith_str)
    if verbose:
        print('Mag Columns Found:\n\t'+'\t'.join(['%s' % c for c in cols]))
    return cols


def convert_deg2rad(df, inplace=False):
    """ Convert any column units from (deg*) to (rad*)
    (e.g. Euler angles and/or gyro measurements)
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    for col in df_out.columns:
        if '(deg' in col:
            print("Converting '%s' readings from (deg*) to (rad*)." % col)
            df_out[col] = np.deg2rad(df_out[col])
            df_out.rename(columns={col: col.replace('(deg', '(rad')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def convert_rad2deg(df, inplace=False):
    """ Convert any column units from (rad*) to (deg*)
    (e.g. Euler angles and/or gyro measurements)
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    for col in df_out.columns:
        if '(rad' in col:
            print("Converting '%s' readings from (rad*) to (deg*)." % col)
            df_out[col] = np.rad2deg(df_out[col])
            df_out.rename(columns={col: col.replace('(rad', '(deg')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def convert_mm2m(df, inplace=False):
    """ Convert any column units from (mm*) to (m*)
    (e.g. Height in (mm) and/or speed (mm/s) measurements)
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    for col in df_out.columns:
        if '(mm' in col:
            print("Converting '%s' readings from (mm*) to (m*)." % col)
            df_out[col] = df_out[col] * mm_to_m
            df_out.rename(columns={col: col.replace('(mm', '(m')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def convert_mG2mps2(df, mG2mps2=9.80665e-3, inplace=False):
    """ Convert any accel columns from (mG) to (m/s^2).
    The accelerometer columns are retrieved using the `get_accel_cols`
    method.
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    cols_accel = get_accel_cols(df_out)
    for col in cols_accel:
        if '(mG)' in col:
            print("Converting '%s' readings from (mG) to (m/s^2)." % col)
            df_out[col] = mG2mps2*df_out[col]
            df_out.rename(columns={col: col.replace('(mG)', '(m/s^2)')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def convert_G2mps2(df, G2mps2=9.80665, inplace=False):
    """ Convert any accel columns from (G) to (m/s^2).
    The accelerometer columns are retrieved using the `get_accel_cols`
    method.
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    cols_accel = get_accel_cols(df_out)
    for col in cols_accel:
        if '(G)' in col:
            print("Converting '%s' readings from (G) to (m/s^2)." % col)
            df_out[col] = G2mps2*df_out[col]
            df_out.rename(columns={col: col.replace('(G)', '(m/s^2)')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def convert_mps2_to_G(df, G_mps2=9.80665, inplace=False):
    """ Convert any accel columns from (m/s^2) tp (G).
    The accelerometer columns are retrieved using the `get_accel_cols`
    method.
    """
    if inplace:
        df_out = df
    else:
        df_out = df.copy()

    cols_accel = get_accel_cols(df_out)
    for col in cols_accel:
        if '(m/s^2)' in col:
            print("Converting '%s' readings from (m/s^2) to (G)." % col)
            df_out[col] = (1.0/G_mps2)*df_out[col]
            df_out.rename(columns={col: col.replace('(m/s^2)', '(G)')},
                          inplace=True)

    if inplace:
        return None
    else:
        return df_out


def read_from_ide(filepath, update_col_names=True):
    """
    Load IDE-format datafile, as saved by enDAQ devices, into
    a dataframe where the column headers are composed of sensor
    axis, and units.

    Inputs
        filepath: string
            path to IDE file (e.g. 'dataset.IDE')
        update_col_names: bool (True)
            Updates column names to match syntax used by other
            default onavdata datasets

    Returns
        dataframe: pandas dataframe
            Dataframe composed of all channles and sub-channels
    """
    ds = idelib.importFile(filepath)

    df_list = []
    # At this point, merging only subset of channels
    # is supported until `ValueError` is investigated.
    for ch in [84, 80]:
        for sch in ds.channels[ch].subchannels:
            str_col = '{}{} ({})'.format(sch.units[0],
                                         sch.axisName,
                                         sch.units[1])
            events = sch.getSession()
            times_us, data = events.arraySlice()
            times_utc = events.session.utcStartTime + times_us/1e6
            # Optional: work with datetime objects
            # times_utc_obj = (
            #     np.datetime64(events.session.utcStartTime, 's')
            #     + times_us.astype('timedelta64[us]')
            # )
            df = pd.DataFrame({str_col: data}, index=times_utc)
            df_list.append(df)
    df = pd.concat(df_list, axis=1)
    ds.close()

    # Reset index and maintain two time columns:
    #   Absolute: UTC time
    #   Relative: `TimeFromStart (s)`
    df.index.name = 'TimeUTC'
    df['TimeFromStart (s)'] = (df.index - df.index[0])
    df.reset_index(inplace=True)

    if update_col_names:
        df.rename(columns={
                  'AccelerationX (g)': 'AccelX (G)',
                  'AccelerationY (g)': 'AccelY (G)',
                  'AccelerationZ (g)': 'AccelZ (G)',
                  'RotationX (deg/s)': 'AngleRateX (deg/s)',
                  'RotationY (deg/s)': 'AngleRateY (deg/s)',
                  'RotationZ (deg/s)': 'AngleRateZ (deg/s)',
                  # 'QuaternionW (q)': 'QuaternionNavToCaseS',
                  # 'QuaternionX (q)': 'QuaternionNavToCaseI',
                  # 'QuaternionY (q)': 'QuaternionNavToCaseJ',
                  # 'QuaternionZ (q)': 'QuaternionNavToCaseK',
                  # 'TemperatureNone (°C)': 'Temperature (C)',
                  # 'TemperatureControl (°C)': 'TemperatureControl (C)'
                  },
                  inplace=True)

    return df



import os
import toml
import pandas as pd
import numpy as np
from collections import OrderedDict
from onavdata.utils.dataframe_utils import read_from_ide
TIME_COL = 'TimeFromStart (s)'

# Load config information, pointing to data sets.
# Append private config, if available.
path_config = os.path.join(os.path.dirname(__file__), 'configs', 'config.toml')
path_private_config = os.path.join(os.path.dirname(__file__),
                                   'configs',
                                   'private-config.toml')
config = toml.load(path_config)
# Make path absolute
path_dataset = config['onavdata-data']['path_parentdir']
path_dataset = os.path.join(os.path.dirname(__file__),
                            path_dataset)
config['onavdata-data']['path_parentdir'] = path_dataset
if os.path.isfile(path_private_config):
    config.update(toml.load(path_private_config))
del path_config
del path_private_config


def add_dataset(name, path):
    """Add existing dataset with shornames

    Inputs
        name: Name of the dataset
        path: Path to dataset which includes the shortnames
    """
    config[name] = dict()
    config[name]['path_parentdir'] = path


def get_shortname_dict(config):
    """
    {key:shortname, value:
        {paths: {dict of paths},
         description: description string}
    }

    """
    shortname_dict = {}
    # print("Data Sets Found:")
    for dataset in config:
        # print('\t', dataset)
        path_dataset = config[dataset]['path_parentdir']
        path_shortnames = os.path.join(path_dataset, 'shortnames.txt')
        shortnames = toml.load(path_shortnames)

        while len(shortnames) > 0:
            sname, paths_dict = shortnames.popitem()

            # Append base path to relative paths inside dataset `shortnames`
            # file.
            for key in ['data', 'meta']:
                # Handle nested lists of paths
                if type(paths_dict[key]) == list:
                    for i in range(len(paths_dict[key])):
                        paths_dict[key][i] = os.path.join(path_dataset,
                                                          paths_dict[key][i])
                else:
                    paths_dict[key] = os.path.join(path_dataset,
                                                   paths_dict[key])
            shortname_dict[sname] = {}
            shortname_dict[sname]['paths'] = paths_dict
            if 'description' in paths_dict:
                shortname_dict[sname]['description'] = paths_dict[
                        'description']

    return shortname_dict


def print_shortnames(description=False):
    """
    Print all shortnames found.  Associated descriptions will be printed if
    `description` set to True.
    """
    shortname_dict = get_shortname_dict(config)
    for sname in sorted(shortname_dict):
        print('\t', sname)
        if description:
            if 'description' in shortname_dict[sname]:
                print(shortname_dict[sname]['description'])
            else:
                print("(No description available)")


def get_data(shortname=None):
    """
    If no shortname is specified, a sample data set will be returned.
    """
    # TODO: this shouldn't be reloaded everytime.  Or maybe it should?
    shortname_dict = get_shortname_dict(config)
    if shortname is None:
        shortname = list(shortname_dict.keys())[0]
        print('get_data(): No shortname specified so choice will be arbitrary.'
              '  Returning: %s' % shortname)

    assert shortname in shortname_dict, 'Requested `shortname` not found.'

    path_meta = shortname_dict[shortname]['paths']['meta']
    meta = toml.load(path_meta, _dict=OrderedDict)

    path_data_list = shortname_dict[shortname]['paths']['data']

    return get_data_direct(path_data_list, meta)


def get_data_direct(path_data_list, meta):
    """
    Load dataset (string or list of paths) using meta
    dictionary supplied.
    """
    if type(path_data_list) == str:
        path_data_list = [path_data_list]

    df_list = []
    for path_data in path_data_list:
        if path_data.upper().endswith('.IDE'):
            # Load IDE format
            df = read_from_ide(path_data)
        else:
            # Assuming CSV format supported by `pd.read_csv` method
            # Check if associated zip-file is available.  If so, use that.
            if os.path.isfile(path_data+'.zip'):
                path_data = path_data+'.zip'

            # Skip lines, if requested (e.g. multi-line header).
            skiprows = None
            if 'skip-lines' in meta:
                skiprows = meta['skip-lines']

            df = pd.read_csv(path_data, skipinitialspace=True, skiprows=skiprows)

        df = apply_meta(df, meta, path_data=path_data)
        df_list.append(df)
    df = pd.concat(df_list, axis=1)

    # Handle duplicate columns.
    cnt_dupcol = df.columns.duplicated().sum()
    if cnt_dupcol > 0:
        print(("Duplicate columns found (%d)."
               "  Taking non-NAN column values.") % cnt_dupcol)
        # For a given row, take the non-na value
        for col in df.columns[df.columns.duplicated()]:
            df[col] = df[col].bfill(axis=1).iloc[:, 0]
        # Drop duplicates
        df = df.loc[:, ~df.columns.duplicated(keep='first')]

    # Reorder columns using 'columns-rename', if specified,
    # while handling that not all columns are part of the rename meta.
    if 'columns-rename' in meta:
        cols = OrderedDict.fromkeys(meta['columns-rename'].values())
        if TIME_COL in cols:
            cols.pop(TIME_COL)
        # Columns not part of the rename are appended to the end.
        cols.update(OrderedDict.fromkeys(df.columns.copy()))
        cols = list(cols.keys())

        df = df[cols]

    return df

def apply_meta(dfin, meta, path_data='None Supplied'):
    """
    Returns a dataframe with meta parameters applied.
    The input dataframe `dfin` is unchanged working with a copy.

    If available, `path_data` is used to print useful status or
    warning messages.
    """
    df = dfin.copy(deep=True)
    # Strip lingering tabs from columns.
    df.columns = [c.lstrip('\t') for c in df.columns]
    df.columns = [c.strip('\t') for c in df.columns]

    # Optionally drop unnamed columns
    if ('drop-unnamed' in meta) and meta['drop-unnamed']:
        cols_unnamed = df.columns.str.contains('unnamed', case=False)
        df.drop(df.columns[cols_unnamed], axis=1, inplace=True)

    # Specify columns to be loaded.
    # Some columns specified to be used or dropped may not be present
    # in the current path_data.  This is handled.
    if ('columns-use' in meta) and ('columns-drop' in meta):
        raise AttributeError("%s meta cannot specify both 'columns-use' "
                             "and 'columns-drop'." % shortname)

    if 'columns-use' in meta:
        colsuse = meta['columns-use']
        if not isinstance(colsuse, list):
            colsuse = [colsuse]
        colsuse = list(set(df.columns) & set(colsuse))
        colsuse.sort()
        df = df[colsuse]

    if 'columns-drop' in meta:
        colsdrop = meta['columns-drop']
        if not isinstance(colsdrop, list):
            colsdrop = [colsdrop]
        colsdrop = list(set(df.columns) & set(colsdrop))
        df = df.drop(columns=colsdrop)

    # Drop rows based on specified value(s).
    if 'drop-rows-where' in meta:
        for key, value in meta['drop-rows-where'].items():
            # Handle both scalar or list of values.
            if not isinstance(value, list):
                value = [value]
            df = df[~df[key].isin(value)]

    # Keep rows based on specified value(s).
    if 'keep-rows-where' in meta:
        for key, value in meta['keep-rows-where'].items():
            # Handle both scalar or list of values.
            if not isinstance(value, list):
                value = [value]
            df = df[df[key].isin(value)]

    if 'keep-rows-gt' in meta:
        for key, value in meta['keep-rows-gt'].items():
            try:
                df = df[df[key].gt(value)]
            except Exception:
                raise ValueError('Unexpected `keep-rows-gt` entry '
                                 'encountered for `[%s]`' % key)

    if 'keep-rows-lt' in meta:
        for key, value in meta['keep-rows-lt'].items():
            try:
                df = df[df[key].lt(value)]
            except Exception:
                raise ValueError('Unexpected `keep-rows-lt` entry '
                                 'encountered for `[%s]`' % key)

    if 'keep-rows-between' in meta:
        for key, value in meta['keep-rows-between'].items():
            try:
                df = df[df[key].between(*value)]
            except Exception:
                raise ValueError('Unexpected `keep-rows-between` entry '
                                 'encountered for `[%s]`' % key)

    # Override default dtypes.
    # The dtype names must be acceptable to pd.DataFrame.astype() function.
    if 'columns-dtype' in meta:
        df = df.astype(dtype=meta['columns-dtype'])

    # Round columns to decimals specified.
    if 'columns-round-decimals' in meta:
        for key, value in meta['columns-round-decimals'].items():
            try:
                df[key] = df[key].round(value)
            except KeyError:
                print(('No matching `columns-round` entry '
                       'encountered for `[{}]`\n'
                       '\tFile: {}').format(
                                        key,
                                        path_data.split(os.path.sep)[-1]))
                continue

    # Subtract offset from columns, if present. Note, this can/should be skipped
    # for multiple-file datasets.
    if 'columns-subtract-offset' in meta:
        for key, value in meta['columns-subtract-offset'].items():
            try:
                df[key] -= float(value)
            except KeyError:
                print(("No matching `columns-subtract-offset` entry found in file.\n"
                       "\columns-subtract-offset.{}\n"
                       "\tFile: {}").format(
                                        key,
                                        path_data.split(os.path.sep)[-1]))
                continue


    # Scale columns, if present.  Note, this can/should be skipped
    # for multiple-file datasets.
    if 'columns-scale' in meta:
        for key, value in meta['columns-scale'].items():
            try:
                df[key] *= float(value)
            except KeyError:
                print(("No matching `columns-scale` entry found in file.\n"
                       "\columns-scale.{}\n"
                       "\tFile: {}").format(
                                        key,
                                        path_data.split(os.path.sep)[-1]))
                continue

    # Rename columns.
    if 'columns-rename' in meta:
        df = df.rename(columns=meta['columns-rename'])

    # Apply Rotations, if present.  Note, this can/should be skipped
    # for multiple-file datasets.
    # NOTE: The rotations are applied AFTER any column renaming.
    #       Hence, the key entries should refer to the renamed column
    #       names.
    if 'Rsensor2body' in meta:
        Rdict = meta['Rsensor2body']
        for key in Rdict:
            cols = [col for col in df if col.startswith(key)]
            if len(cols) == 0:
                print(("No matching Rsensor2body columns found in file.\n"
                       "\tRsensor2body.{}\n"
                       "\tFile: {}").format(
                                        key,
                                        path_data.split(os.path.sep)[-1]))
                continue
            # Sort the columns so that we end up with the order: X->Y->Z
            # where these are in the `sensor` frame.
            cols.sort()
            assert len(cols) == 3, ("Invalid number of matching columns "
                                    "found for Rsensor2body.{}:\n"
                                    "cols found: {}").format(key, cols)

            # Apply transformation:
            #   [x1, x2, ... xm]                   [x1, x2, ... xm]
            #   [y1, y2, ... ym]  =   Rsensor2body [y1, y2, ... ym]
            #   [z1, y3, ... ym]                   [z1, y3, ... ym]
            #                   body                               sensor
            # We transpose the entire expression in order to be able
            # to work with and store the Pandas DataFrame directly.
            Rsensor2body = np.array(Rdict[key], dtype=float)
            df[cols] = df[cols] @ Rsensor2body.transpose()

    # Check for time column and attempt to populate if it doesn't exist.
    if df.index.name != TIME_COL:
        if TIME_COL not in df:
            if 'SampleFreq (Hz)' in meta:
                dt_sec = 1.0/meta['SampleFreq (Hz)']

                df[TIME_COL] = df.index * dt_sec
            else:
                print("Unable to index time information.")
        df.set_index(TIME_COL, inplace=True)

    # Remove rows with duplicate index.
    cnt_dupind = df.index.duplicated().sum()
    if cnt_dupind > 0:
        print("Duplicate indices found (%d).  Keeping first." % cnt_dupind)
        df = df[~df.index.duplicated(keep='first')]

    # Add any columns specified via meta file.
    if 'add-columns' in meta:
        for key, value in meta['add-columns'].items():
            try:
                df[key] = df[value].copy()
            except KeyError:
                df[key] = value
            except Exception:
                raise ValueError('Unexpected `add-columns` entry '
                                 'encountered for `[%s]`' % key)
    return df


if __name__ == '__main__':
    print("Data Shortnames Found:")
    print_shortnames(description=True)



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="onavdata",
    version="0.2.2",
    author="Organic Navigation",
    author_email="team@organicnavigation.com",
    description=(("Easily import both reference and user-specified datasets "
                  "for navigation system design and testing projects.")),
    packages=['onavdata', 'onavdata.utils'],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    scripts=['bin/onavdata-print-shortnames'],
    install_requires=[
          'toml',
          'pandas',
          'numpy',
          'idelib==3.1.0'
      ],
)



from onavdata.utils.dataframe_utils import get_accel_cols
from onavdata.utils.dataframe_utils import get_gyro_cols
from onavdata.utils.dataframe_utils import get_mag_cols
from onavdata.utils.dataframe_utils import convert_deg2rad
from onavdata.utils.dataframe_utils import convert_rad2deg
from onavdata.utils.dataframe_utils import convert_mG2mps2
from onavdata.utils.dataframe_utils import convert_G2mps2
from onavdata.utils.dataframe_utils import convert_mps2_to_G
from onavdata.utils.dataframe_utils import convert_mm2m
